package com.jniwrapper.win32.hook;

import com.jniwrapper.*;
import com.jniwrapper.win32.FunctionName;
import com.jniwrapper.win32.Handle;
import com.jniwrapper.win32.MessageLoopThread;
import com.jniwrapper.win32.system.Module;
import com.jniwrapper.win32.ui.User32;

/**
 * Provides base implementation of a low-level hook.
 *
 * @author Serge Piletsky
 */
abstract class LowLevelHook extends Hook
{
    private final Handle _hookHandle = new Handle();
    private MessageLoopThread _dispatchThread;
    private HookProc _callback;

    LowLevelHook(Descriptor descriptor)
    {
        super(descriptor);
        _dispatchThread = new MessageLoopThread("HookDispatchThread");
    }

    public boolean isInstalled()
    {
        return _dispatchThread.isStarted();
    }

    public void install()
    {
        _dispatchThread.doStart();
        _callback = new HookProc(this);
        // install a hook in a dispatch thread
        try
        {
            _dispatchThread.doInvokeAndWait(new Runnable()
            {
                public void run()
                {
                    User32 user32 = User32.getInstance();
                    Function setWindowsHook = user32.getFunction(new FunctionName("SetWindowsHookEx").toString());
                    UInt32 threadID = new UInt32();

                    final Module currentModule = Module.getCurrent();
                    setWindowsHook.invoke(_hookHandle, new Parameter[]{
                            new Int(getDescriptor().getValue()),
                            _callback,
                            currentModule,
                            threadID
                    });
                }
            });
        }
        catch (Exception e)
        {
            throw new RuntimeException("Failed to install low level hook", e);
        }
    }

    public void uninstall()
    {
        try
        {
            _dispatchThread.doInvokeAndWait(new Runnable()
            {
                public void run()
                {
                    User32 user32 = User32.getInstance();
                    Function unhook = user32.getFunction("UnhookWindowsHookEx");
                    unhook.invoke(null, _hookHandle);
                }
            });
        }
        catch (Exception e)
        {
            throw new RuntimeException("Failed to uninstall low level hook", e);
        }
        finally
        {
            _dispatchThread.doStop();
            _callback.dispose();
        }
    }

    protected abstract long processEvent(long wParam, long lParam);

    private static class HookProc extends Callback
    {
        private Int code = new Int();
        private UInt32 wParam = new UInt32();
        private UInt32 lParam = new UInt32();
        private UInt32 lResult = new UInt32();

        private LowLevelHook hook;

        private static Function CallNextHookEx = User32.getInstance().getFunction("CallNextHookEx");

        public HookProc(LowLevelHook hook)
        {
            if (hook == null)
            {
                throw new NullPointerException();
            }

            this.hook = hook;
            init(new Parameter[]{code, wParam, lParam}, lResult);
        }

        public void callback()
        {
            if (code.getValue() < 0)
            {
                CallNextHookEx.invoke(lResult, hook._hookHandle, code, wParam, lParam);
            }
            else
            {
                long nCode = hook.processEvent(wParam.getValue(), lParam.getValue());
                if (nCode >= 0)
                {
                    CallNextHookEx.invoke(lResult, hook._hookHandle, code, wParam, lParam);
                }
                else
                {
                    lResult.setValue(nCode);
                }
            }
        }
    }

    public void setSynchronous(boolean synchronous)
    {
        // it always work in synchronous mode
    }

    public boolean isSynchronous()
    {
        return true;
    }
}
